package dataWrangler;

/*
 * ThingComparer.java
 *
 * Created on March 14, 2003, 7:49 PM
 */

/**
 *
 * @author  rod
 */
import java.util.*;

public class ThingComparer {
    
    /* type identifiers */
    public static final int INT = 1;
    public static final int LONG = 2;
    public static final int STRING = 3;
    
    /* comparison methods */
    public static final int STRING_COMPARE_FIRST = 0;
    public static final int STRING_COMPARE_CONTAINS = 1;
    public static final int NUM_COMPARE_LESS = 2;
    public static final int NUM_COMPARE_EQUAL = 3;
    public static final int NUM_COMPARE_GREATER = 4;
    public static final int NUM_COMPARE_GAUSSIAN = 5;
    public static final int NUM_COMPARE_ABS_DIFF = 6;
    public static final int STRING_COMPARE_EQUALS = 7;
    public static final int STRING_COMPARE_SHARED_WORDS = 8;
    public static final int STRING_COMPARE_REGEX = 9;
    
    public static final String[] comparisons = {
        "begins with",
        "contains",
        "is less than",
        "equals",
        "is greater than",
        "is approximately (1)",
        "is approximately (2)",
        "equals",
        "has words",
        "matches regex"
    };
    
    public static final int[] comparisonTransition = {
        1, 7, 3, 4, 5, 6, 2, 8, 9, 0
    };
        
    /** Creates a new instance of ThingComparer */
    public ThingComparer() {
    }
    
    /* focused object is Thing One */
    public static Vector compareToAll(Thing t1, Vector tv, int dim, int method){
if (t1 == null) return null;
        switch(Thing.dimensionTypes[dim]){
            case INT: return compareAll(t1, tv, dim, new IntComparer(method));
            case LONG: return compareAll(t1, tv, dim, new LongComparer(method));
            case STRING:  return compareAll(t1, tv, dim, new StringComparer(method));
        }
        return new Vector(); // or null...
    }
    
    private static Vector compareAll(Thing t, Vector tv, int dim, Comparer cmp){
        int tIndex = tv.indexOf(t);
        Vector result = new Vector();
        for (int i = 0; i < tIndex; i++){
if(tv.elementAt(i)==null) result.addElement(new Double(-1));
else 
            result.addElement(cmp.compare(t, (Thing)(tv.elementAt(i)), dim));
        }
        result.addElement(new Double(-1)); // skip self
        for (int i = tIndex+1; i < tv.size(); i++){
if(tv.elementAt(i)==null) result.addElement(new Double(-1));
else 
            result.addElement(cmp.compare(t, (Thing)(tv.elementAt(i)), dim));
        }
        return result;
    }
}

interface Comparer{
    public Double compare(Thing reference, Thing thing, int dim);
}

class StringComparer implements Comparer{
    
    int method;
    
    public StringComparer(int method){
        this.method = method;
    }
    
    public Double compare(Thing t1, Thing t2, int dim){
        double res = 1;
if ((t1==null)||(t2==null)||(t1.getByDimension(dim)==null)||(t2.getByDimension(dim)==null))
return new Double(-1);
        String s1 = t1.getByDimension(dim).toLowerCase();
        String s2 = t2.getByDimension(dim).toLowerCase();
        if ((s1 != null) && (s2 != null)){
        switch (method){
            case ThingComparer.STRING_COMPARE_EQUALS:
                if(s1.equals(s2)) res=0; break;
            case ThingComparer.STRING_COMPARE_SHARED_WORDS:
                res = sharedWordCompare(s1, s2); break;
            case ThingComparer.STRING_COMPARE_FIRST:
                if(s2.startsWith(s1)) res = 0; break;
            case ThingComparer.STRING_COMPARE_CONTAINS:
                if(s2.indexOf(s1)!=-1) res = 0; break;
            case ThingComparer.STRING_COMPARE_REGEX:
                if(s2.matches(s1)) res = 0; break;
	    default:System.out.println("!");
        }
        }
        return new Double(res);
    }
    
    private double sharedWordCompare(String s1, String s2){
        int matches = 0;
        HashSet h1 = wordsIn(s1);
        HashSet h2 = wordsIn(s2);
        Iterator i = h1.iterator();
        while(i.hasNext()){
            if(h2.contains(i.next()))
                matches++;
        }
        return 1.0-(Math.max((double)matches/(double)(h1.size()),1));
        
    }
    
    private HashSet wordsIn(String s){
        HashSet hs = new HashSet();
        String word="";
        for (int i = 0; i<s.length(); i++){
            if(!Character.isLetter(s.charAt(i))){
                if(!word.equals("")){
                    hs.add(word);
                    word="";
                }
            }else{
                word += s.charAt(i);
            }
        }
        if (!word.equals(""))
            hs.add(word);
        return hs;
    }
}


class IntComparer implements Comparer{
    /* just using the same methods as for long */
    int method;
    LongComparer lc;
    
    public IntComparer(int method){
        lc = new LongComparer(method);
        this.method = method;
    }
    
    public Double compare(Thing t1, Thing t2, int dim){
        return lc.compare(t1, t2, dim);
    }
}

class LongComparer implements Comparer{
    
    int method;
    
    public LongComparer(int method){
        this.method = method;
    }
    
    public Double compare(Thing t1, Thing t2, int dim){
        double res = 1;
        long L1 = Long.parseLong(t1.getByDimension(dim));
        long L2 = Long.parseLong(t2.getByDimension(dim));
        switch (method){
            case ThingComparer.NUM_COMPARE_LESS:
                if(L2<L1) res=0; break;
            case ThingComparer.NUM_COMPARE_EQUAL:
                if(L2==L1) res=0; break;
            case ThingComparer.NUM_COMPARE_GREATER:
                if(L2>L1) res = 0; break;
            case ThingComparer.NUM_COMPARE_GAUSSIAN:
                res = gaussian(L1, L2); break;
            case ThingComparer.NUM_COMPARE_ABS_DIFF:
                res = absDiff(L1, L2); break;
        }
        return new Double(res);
    }
    
    private static double gaussian(double mean, double sample){
        //final double c = 1.0 / Math.sqrt(2.0 * Math.PI);
        double result = /* c * */ Math.exp(-0.5*((sample - mean)*(sample - mean)));
        return 1-result;
    }
    
    private static double absDiff(double target, double sample){
        return Math.abs(sample - target)/(Math.abs(sample)+Math.abs(target));
    }
    
}
